# -*- coding: utf-8 -*-
"""
==============================================================================
@author: Nikolaos Giakoumoglou
@date: Thu May 20 13:46:57 2021
@reference: Teague, Image analysis via the general theory of moments
==============================================================================
"""

import mahotas

def zernikes_moments(f, radius=9):
    '''
    Parameters
    ----------
    f : numpy ndarray
        Image of dimensions N1 x N2.
    radius : int, optional
        Radius to calculate Zernikes moments. The default is 9.

    Returns
    -------
    features : numpy ndarray
        Zernikes' moments.
    labels : list
        Labels of features.
    '''
    
    features = mahotas.features.zernike_moments(f, radius)
    labels = ['Zernikes_Moments_radius_' + str(radius) + '_' + str(i) for i in range(features.shape[0])]
    return features, labels


# -*- coding: utf-8 -*-
"""
==============================================================================
@author: Nikolaos Giakoumoglou
@date: Thu May 13 12:14:15 2021
@reference: Tsiaparas, Comparison of Multiresolution Features for Texture Classification of Carotid Atherosclerosis From B-Mode Ultrasound
==============================================================================
"""

import pywt
import numpy as np

def wp_features(f, mask, wavelet='coif1', maxlevel=3): 
    ''' 
    Parameters
    ----------
    f : numpy ndarray
        Image of dimensions N1 x N2.
    mask : numpy ndarray
        Mask image N1 x N2 with 1 if pixels belongs to ROI, 0 else. Give None
        if you want to consider ROI the whole image.
    wavelet : str, optional
         Filter to be used. Check pywt for filter families. The default is 'cof1'
    maxlevel : int, optional
        Levels of decomposition. Default is 3.

    Returns
    -------
    features : numpy ndarray
        Mean and std of each detail image. Appromimation images are ignored.
    labels : list
        Labels of features.
    '''
    
    if mask is None:
        mask = np.ones(f.shape)
        
    # Step 1: Get Wavelet Decomposition in 3 levels 
    wp = pywt.WaveletPacket2D(data=f, wavelet=wavelet, mode='symmetric')
    wp_mask = pywt.WaveletPacket2D(data=mask, wavelet=wavelet, mode='symmetric')
    
    # Step 2: Check if we exceed max levels
    if maxlevel > wp.maxlevel:
        maxlevel = wp.maxlevel

    # Step 3: For each coeff array (64-1 sub-images), get mean and std
    features = np.zeros((len(wp.get_level(maxlevel))-1,2), np.double)
    paths = [node.path for node in wp.get_level(maxlevel, 'natural')]
    paths.remove('a' * maxlevel) # remove top left array
    labels = []
    for i in range(len(paths)):
        D_f = wp[paths[i]].data
        D_mask = wp_mask[paths[i]].data
        D_mask[D_mask != 0] = 1
        D = D_f.flatten()[D_mask.flatten().astype(bool)] # work on elements inside mask
        features[i][0], features[i][1] = abs(D).mean(), abs(D).std()
        labels.append('WP_' + str(wavelet) + '_' + str(paths[i]) + '_mean')
        labels.append('WP_' + str(wavelet) + '_' + str(paths[i]) + '_std')
    
    # Step 4: Return
    return features.flatten(), labels


# -*- coding: utf-8 -*-
"""
==============================================================================
@author: Nikolaos Giakoumoglou
@date: Mon May 10 12:42:30 2021
@reference: Maragos, Pattern Spectrum and Multiscale Shape Representation
            Maragos, Threshold Superposition in Morphological Image Analysis Systems
==============================================================================
"""

import numpy as np
import matplotlib.pyplot as plt
from skimage import morphology

def _opening_FP(f, g, n): # (f o ng), n=0,1,2... 
    ''' 
    Parameters
    ----------
    f : numpy ndarray
        Image of dimensions N1 x N2.
    g : numpy ndarray
        Structural element/pattern/kernel
    n : np.array
        Number of scales.

    Returns
    -------
    out : numpy ndarray
        Multiscale function-processing opening (f o ng) at scale n.
    '''
    
    out = f.copy()
    for i in range(n):
        out = morphology.erosion(out, g)
    for i in range(n):
        out = morphology.dilation(out,g)
    return out 

def _pattern_spectrum(f, g, n):
    ''' 
    Parameters
    ----------
    f : numpy ndarray
        Image of dimensions N1 x N2.
    g : numpy ndarray
        Structural element/pattern/kernel
    n : np.array
        Number of scales.

    Returns
    -------
    out : numpy ndarray
        Pattern spectrum PS(f,g,n) = A[f o ng - f o (n+1)g] 
    '''
    ps = _opening_FP(f,g,n) - _opening_FP(f,g,(n+1))
    return ps.sum() 

def grayscale_morphology_features(f, N=30):
    ''' 
    Parameters
    ----------
    f : numpy ndarray
        Image of dimensions N1 x N2.
    N : np.array, optional
        Maximum number of scales. The default is 30.

    Returns
    -------
    pdf : numpy ndarray
        Probability density function (pdf) of pattern spectrum.
    cdf : numpy ndarray
        Cumulative density function (cdf) of pattern spectrum.
    '''
    f = f.astype(np.uint8)                # grayscale image
    kernel = np.ones((3,3), np.uint8)     # kerne: cross '+'
    kernel[0,0], kernel[2,2], kernel[0,2], kernel[2,0] = 0, 0, 0, 0 
    ps = np.zeros(N, np.double)           # pattern spectrum
    for n in range(N):
        ps[n] = _pattern_spectrum(f,kernel,n)
    pdf = ps / f.sum() 
    cdf = np.cumsum(pdf)  
    return pdf, cdf

def plot_pdf_cdf(pdf, cdf, name=''):
    if name != '':
        name = '('+name+')'
    fig, (ax1, ax2) = plt.subplots(1,2)
    fig.suptitle('Plaque Grayscale Morphological Features ' + str(name))
    ax1.plot(pdf)
    ax1.set_title('pdf')
    ax2.plot(cdf)
    ax2.set_title('cdf')


from .hog import hog_features, plot_hog
from .hu import hu_moments
from .tas import tas_features
from .zernikes import zernikes_moments

__all__ = ['hog_features', 'plot_hog',
           'hu_moments',
           'tas_features',
           'zernikes_moments']


# -*- coding: utf-8 -*-
"""
==============================================================================
@author: Nikolaos Giakoumoglou
@date: Thu May  6 19:10:11 2021
@reference: Haralick, Textural Features for Image Classification 
==============================================================================
"""
import numpy as np
import mahotas

def glcm_features(f, ignore_zeros=True):
    '''
    Parameters
    ----------
    f : numpy ndarray
        Image of dimensions N1 x N2.
    ignore_zeros : int, optional
        Ignore zeros in image f. The default is True.

    Returns
    -------
    features_mean : numpy ndarray
        Haralick's 1)Angular Second Moment, 2)Contrast, 
        3)Correlation, 4)Sum of Squares: Variance, 5)Inverse 
        Difference Moment 6)Sum Average, 7)Sum Variance, 8)Sum 
        Entropy, 9)Entropy, 10)Difference Variance, 11)Difference 
        Entropy, 12)Information Measure of Correlation 1, 
        13)Information Measure of Correlation 2, 14)Maximal 
        Correlation Coefficient, mean
    features_range : numpy ndarray
        Haralick's features, same as before but range
    labels_mean : list
        Labels of features_mean.
    labels_range: list
        Labels of features_range.
    '''
       
    # 1) Labels
    labels = ["GLCM_ASM", "GLCM_Contrast", "GLCM_Correlation",
              "GLCM_SumOfSquaresVariance", "GLCM_InverseDifferenceMoment",
               "GLCM_SumAverage", "GLCM_SumVariance", "GLCM_SumEntropy",
               "GLCM_Entropy", "GLCM_DifferenceVariance",
               "GLCM_DifferenceEntropy", "GLCM_Information1",
               "GLCM_Information2", "GLCM_MaximalCorrelationCoefficient"]
    labels_mean = [label + "_Mean" for label in labels]
    labels_range = [label + "_Range" for label in labels]
    
    # 2) Parameters
    f = f.astype(np.uint8)
    
    # 3) Calculate Features: Mean and Range
    features = mahotas.features.haralick(f, 
                                         ignore_zeros=ignore_zeros, 
                                         compute_14th_feature=True,
                                         return_mean_ptp=True)
    features_mean = features[0:14]
    features_range = features[14:]
    
    return features_mean, features_range, labels_mean, labels_range



# -*- coding: utf-8 -*-
"""
==============================================================================
@author: Nikolaos Giakoumoglou
@date: May  8 15:18:53 2021
@reference: Weszka, A Comparative Study of Texture Measures for Terrain Classification
            Wu, Texture Features for Classification
==============================================================================
"""

import numpy as np

# FIXME
def _fft2(x,msk):
    '''
    Parameters
    ----------
    f : numpy ndarray
        Image of dimensions N1 x N2.
    msk : numpy ndarray
        Mask image N1 x N2 with 1 if pixels belongs to ROI, 0 else.

    Returns
    -------
    F : numpy ndarray
        2D Fourier Transform inside msk.
    '''
    N1,N2 = x.shape
    F = np.zeros((N1,N2), np.complex64)
    
    for i in range(1,N2+1):
      sp = 0
      ep = 0
      j = 1
      while (j < N1):
          while (msk[j-1,i-1]==1) & (j < N1):
              if (sp == 0):
                  sp = j
              j += 1
          if (sp > 0) & (ep == 0):
              if (j < N1):
                  ep = j-1
              else:
                  ep = j      
              F[(sp-1):(ep),i-1] = np.fft.fft(x[(sp-1):(ep),i-1])
              sp = 0
              ep = 0
          while (msk[j-1,i-1] == 0) & (j < N1):
              j += 1

    F = F.T
    msk = msk.T
    
    for i in range(1,N1+1):
        sp = 0
        ep = 0
        j = 1
        while (j < N2):
          while (msk[j-1,i-1] == 1) & (j < N2):
              if (sp == 0):
                  sp = j
              j += 1
          if (sp > 0) & (ep == 0):
              if (j < N1):
                  ep = j-1
              else:
                  ep = j     
              F[(sp-1):(ep),i-1] = np.fft.fft(F[(sp-1):(ep),i-1])
              sp = 0
              ep = 0
          while (msk[j-1,i-1] == 0) & (j < N2):
              j += 1

    F = F.T
    msk = msk.T
    return F

def fps(f, mask):
    '''
    Parameters
    ----------
    f : numpy ndarray
        Image of dimensions N1 x N2.
    mask : numpy ndarray
        Mask image N1 x N2 with 1 if pixels belongs to ROI, 0 else. Give None
        if you want to consider ROI the whole image.

    Returns
    -------
    features : numpy ndarray
        1)Radial Sum, 2)Angular Sum.
    labels : list
        Labels of features.
    '''

    if mask is None:
        mask = np.ones(f.shape)
        
    # 1) Labels
    labels = ["FPS_RadialSum", "FPS_AngularSum"]
    
    # 2) Parameters
    f = np.array(f, np.double)
    mask = np.array(mask, np.double)
     
    # 3) Calculate Features
    area = mask.sum()
    F = _fft2(f, mask)
    F_real = np.real(F)
    F_imag = np.imag(F)
    #F_real = np.multiply(F_real, mask)
    #F_imag = np.multiply(F_imag, mask)
    features = np.zeros(2 ,np.double)
    features[0] = np.sqrt(sum(sum(np.multiply(F_real,F_real)))/area)
    features[1] = np.sqrt(sum(sum(np.multiply(F_imag,F_imag)))/area)
    return features, labels
 



# -*- coding: utf-8 -*-
"""
==============================================================================
@author: Nikolaos Giakoumoglou
@author: https://github.com/eiproject/lib-GLRLM-Python3/blob/master/lib-GLRLM/libpreprocessing.py
@author: https://github.com/szhHr/Gray-Level-Run-Length-Matrix-to-get-image-feature/blob/master/GrayRumatrix.py
@date: Sat May  8 17:00:35 2021
@reference: Gallowway, Texture Analysis using Gray Level Run Lengths
==============================================================================
"""
    
import numpy as np

def glrlm_0(f, mask, grayLevel=5, runLength=5, skipFirstRow=True):
    degree0Matrix = np.zeros([grayLevel, runLength])
    counter = 0
    for y in range(f.shape[0]):
        for x in range(f.shape[1]):   
            nowVal = f[y][x]
            if x + 1 >= f.shape[1]:
                nextVal = None
            else:
                nextVal = f[y][x + 1]
            if nextVal != nowVal and counter == 0:
                degree0Matrix[int(nowVal)][counter] += 1
            elif nextVal == nowVal:
                counter += 1
            elif nextVal != nowVal and counter != 0:
                degree0Matrix[int(nowVal)][counter] += 1
                counter = 0
    return degree0Matrix[1:,:] if skipFirstRow else degree0Matrix

def glrlm_90(f, mask, grayLevel=5, runLength=5, skipFirstRow=True):
    degree90Matrix = np.zeros([grayLevel, runLength])
    counter = 0
    for x in range(f.shape[1]):
        for y in range(f.shape[0]):
            nowVal = f[y][x]
            if y + 1 >= f.shape[0]:
                nextVal = None
            else:
                nextVal = f[y + 1][x]
            if nextVal != nowVal and counter == 0:
                degree90Matrix[int(nowVal)][counter] += 1
            elif nextVal == nowVal:
                counter += 1
            elif nextVal != nowVal and counter != 0:
                degree90Matrix[int(nowVal)][counter] += 1
                counter = 0
    return degree90Matrix[1:,:] if skipFirstRow else degree90Matrix

def glrlm_45(f, mask, grayLevel=5, runLength=5, skipFirstRow=True):
    degree45Matrix = np.zeros([grayLevel, runLength])
    for y in range(f.shape[0]):
        counter = 0
        i_range = max(f.shape)
        for i in range(i_range):
            y1 = y - i
            if i >= f.shape[1] or y1 < 0:
                break
            else:
                nowVal = f[y1][i]
            if y1 - 1 < 0 or i + 1 >= f.shape[1]:
                nextVal = None
            else:
                nextVal = f[y1 - 1][i + 1]
            if nextVal != nowVal and counter == 0:
                degree45Matrix[int(nowVal)][counter] += 1
            elif nextVal == nowVal:
                counter += 1
            elif nextVal != nowVal and counter != 0:
                degree45Matrix[int(nowVal)][counter] += 1
                counter = 0
    for x in range(f.shape[1]):
        if x == f.shape[1] - 1:
            break
        counter = 0
        i_range = max(f.shape)
        for i in range(i_range):
            y_i = -1 - i
            x_i = -1 + i - x
            if x_i >= 0 or y_i <= -1 - f.shape[0]:
                break
            else:
                nowVal = f[y_i][x_i]
            if y_i - 1 <= -(f.shape[0] + 1) or x_i + 1 >= 0:
                nextVal = None
            else:
                nextVal = f[y_i - 1][x_i + 1]
            if nextVal != nowVal and counter == 0:
                degree45Matrix[int(nowVal)][counter] += 1
            elif nextVal == nowVal:
                counter += 1
            elif nextVal != nowVal and counter != 0:
                degree45Matrix[int(nowVal)][counter] += 1
                counter = 0
    degree45Matrix[0,1:] = 0
    return degree45Matrix[1:,:] if skipFirstRow else degree45Matrix

def glrlm_135(f, mask, grayLevel=5, runLength=5, skipFirstRow=True):
    degree135Matrix = np.zeros([grayLevel, runLength])
    for y in range(f.shape[0]):
        counter = 0
        i_range = max(f.shape)
        for i in range(i_range):
            y1 = y + i
            if y1 >= f.shape[0] or i >= f.shape[1]:
                break
            else:
                nowVal = f[y1][i]
                if y1 >= f.shape[0] - 1 or i >= f.shape[1] - 1:
                    nextVal = None
                else:
                    nextVal = f[y1 + 1][i + 1]
                if nextVal != nowVal and counter == 0:
                    degree135Matrix[int(nowVal)][counter] += 1
                elif nextVal == nowVal:
                    counter += 1
                elif nextVal != nowVal and counter != 0:
                    degree135Matrix[int(nowVal)][counter] += 1
                    counter = 0
    for x in range(f.shape[1]):
        if x == 0:
            continue
        i_range = max(f.shape)
        counter = 0
        for i in range(i_range):
            x1 = x + i
            if i >= f.shape[0] or x1 >= f.shape[1]:
                break
            else:
                nowVal = f[i][x1]
            if i >= f.shape[0] - 1 or x1 >= f.shape[1] - 1:
                nextVal = None
            else:
                nextVal = f[i + 1][x1 + 1]
            if nextVal != nowVal and counter == 0:
                degree135Matrix[int(nowVal)][counter] += 1
            elif nextVal == nowVal:
                counter += 1
            elif nextVal != nowVal and counter != 0:
                degree135Matrix[int(nowVal)][counter] += 1
                counter = 0
    degree135Matrix[0,1:] = 0
    return degree135Matrix[1:,:] if skipFirstRow else degree135Matrix

def _apply_over_degree(function, x1, x2):
    if function == np.divide:
        x2 = x2 + 1e-16
    rows, cols, nums = x1.shape
    result = np.ndarray((rows, cols, nums))
    for i in range(nums):
        result[:, :, i] = function(x1[:, :, i], x2)
        result[result == np.inf] = 0
        result[np.isnan(result)] = 0
    return result 
    
def _calculate_ij (rlmatrix):
    gray_level, run_length, _ = rlmatrix.shape
    I, J = np.ogrid[0:gray_level, 0:run_length]
    return I, J+1
    
def _calculate_s(rlmatrix):
    return np.apply_over_axes(np.sum, rlmatrix, axes=(0, 1))[0, 0]

def glrlm(f, mask, Ng=256):   
    '''
    Parameters
    ----------
    f : numpy ndarray
        Image of dimensions N1 x N2.
    mask : numpy ndarray
        Mask image N1 x N2 with 1 if pixels belongs to ROI, 0 else.
    Ng : int, optional
        Image number of gray values. The default is 256.

    Returns
    -------
    mat : numpy ndarray
        GLRL Matrices for 0, 45, 90 and 135 degrees.
    '''
    runLength = max(f.shape)
    mat0 = glrlm_0(f, mask, grayLevel=Ng, runLength=runLength)
    mat45 = glrlm_45(f, mask, grayLevel=Ng, runLength=runLength)
    mat90 = glrlm_90(f, mask, grayLevel=Ng, runLength=runLength)
    mat135 = glrlm_135(f, mask, grayLevel=Ng, runLength=runLength)            
    mat = np.dstack((mat0, mat45, mat90, mat135))      
    return mat

def glrlm_features(f, mask, Ng=256):
    '''
    Parameters
    ----------
    f : numpy ndarray
        Image of dimensions N1 x N2.
    mask : numpy ndarray
        Mask image N1 x N2 with 1 if pixels belongs to ROI, 0 else. Give None
        if you want to consider ROI the whole image.
    Ng : int, optional
        Image number of gray values. The default is 256.

    Returns
    -------
    features : numpy ndarray
        1)Short Run Emphasis, 2)Long Run Emphasis, 3)Gray Level 
        Non-Uniformity/Gray Level Distribution, 4)Run Length 
        Non-Uniformity/Run Length Distribution, 5)Run Percentage,
        6)Low Gray Level Run Emphasis, 7)High Gray Level Run Emphasis,
        8)Short Low Gray Level Emphasis, 9)Short Run High Gray Level 
        Emphasis, 10)Long Run Low Gray Level Emphasis, 11)Long Run 
        High Gray Level Emphasis.
    labels : list
        Labels of features.
    '''
    
    if mask is None:
        mask = np.ones(f.shape)
        
    labels = ["GLRLM_ShortRunEmphasis",
              "GLRLM_LongRunEmphasis",
              "GLRLM_GrayLevelNo-Uniformity",
              "GLRLM_RunLengthNonUniformity",
              "GLRLM_RunPercentage",
              "GLRLM_LowGrayLevelRunEmphasis",
              "GLRLM_HighGrayLevelRunEmphasis",
              "GLRLM_Short owGrayLevelEmphasis",
              "GLRLM_ShortRunHighGrayLevelEmphasis",
              "GLRLM_LongRunLowGrayLevelEmphasis",
              "GLRLM_LongRunHighGrayLevelEmphasis"]
    
    rlmatrix = glrlm(f, mask, Ng)
        
    I, J = _calculate_ij(rlmatrix)
    S = _calculate_s(rlmatrix)
    G = np.apply_over_axes(np.sum, rlmatrix, axes=1)
    R = np.apply_over_axes(np.sum, rlmatrix, axes=0)
        
    features = np.zeros(11,np.double)
    features[0] = ((np.apply_over_axes(np.sum, _apply_over_degree(np.divide, rlmatrix, (J*J)), axes=(0, 1))[0, 0])/S).mean()
    features[1] = ((np.apply_over_axes(np.sum, _apply_over_degree(np.multiply, rlmatrix, (J*J)), axes=(0, 1))[0, 0])/S).mean()
    features[2] = ((np.apply_over_axes(np.sum, (G*G), axes=(0, 1))[0, 0])/S).mean()
    features[3] = ((np.apply_over_axes(np.sum, (R*R), axes=(0, 1))[0, 0])/S).mean()
        
    gray_level, run_length,_ = rlmatrix.shape
    num_voxels = gray_level * run_length
    features[4] = (S/num_voxels).mean()
        
    features[5]= ((np.apply_over_axes(np.sum, _apply_over_degree(np.divide, rlmatrix, (I*I)), axes=(0, 1))[0, 0])/S).mean()
    features[6] = ((np.apply_over_axes(np.sum, _apply_over_degree(np.multiply, rlmatrix, (I*I)), axes=(0, 1))[0, 0])/S).mean()
    features[7] = ((np.apply_over_axes(np.sum, _apply_over_degree(np.divide, rlmatrix, (I*I*J*J)), axes=(0, 1))[0, 0])/S).mean()
        
    temp = _apply_over_degree(np.multiply, rlmatrix, (I*I))
    features[8] = ((np.apply_over_axes(np.sum, _apply_over_degree(np.divide, temp, (J*J)), axes=(0, 1))[0, 0])/S).mean()
        
    temp = _apply_over_degree(np.multiply, rlmatrix, (J*J))
    features[9] = ((np.apply_over_axes(np.sum, _apply_over_degree(np.divide, temp, (J*J)), axes=(0, 1))[0, 0])/S).mean()
    features[10] = ((np.apply_over_axes(np.sum, _apply_over_degree(np.multiply, rlmatrix, (I*I*J*J)), axes=(0, 1))[0, 0])/S).mean()
        
    return features, labels    


# -*- coding: utf-8 -*-
"""
==============================================================================
@author: Nikolaos Giakoumoglou
@date: Fri May  7 19:28:12 2021
@reference: Wu, Texture Features for Classification
            Law, Rapid Texture Identification
            Haralick, Computer and Robot Vision Vol. 1
==============================================================================
"""

import numpy as np
from scipy import signal
from ..utilities import _image_xor

def lte_measures(f, mask, l=7):
    '''
    Parameters
    ----------
    f : numpy ndarray
        Image of dimensions N1 x N2.
    mask : numpy ndarray
        Mask image N1 x N2 with 1 if pixels belongs to ROI, 0 else. Give None
        if you want to consider ROI the whole image.
    l : int, optional
        Law's mask size. The default is 7.

    Returns
    -------
    features : numpy ndarray
        1)texture energy from LL kernel, 2) texture energy from EE 
        kernel, 3)texture energy from SS kernel, 4)average texture 
        energy from LE and EL kernels, 5)average texture energy from 
        ES and SE kernels, 6)average texture energy from LS and SL 
        kernels.
    labels : list
        Labels of features.
    '''

    if mask is None:
        mask = np.ones(f.shape)
        
    # 1) Labels
    labels = ["LTE_LL","LTE_EE","LTE_SS","LTE_LE","LTE_ES","LTE_LS"]
    labels = [label+'_'+str(l) for label in labels]
    
    # 2) Parameters
    f = np.array(f, np.double)
    mask = np.array(mask, np.double) 
    kernels = np.zeros((l,l,9), np.double)
    
    # 3) From 3 kernels [L, E, S], get 9 [LL, LE, LS, EL, EE, ES, SL, SE, SS]
    if l==3:
        L = np.array([ 1,  2,  1], np.double)
        E = np.array([-1,  0,  1], np.double)
        S = np.array([-1,  2, -1], np.double)
    elif l==5:
        L = np.array([ 1,  4,  6,  4,  1], np.double)
        E = np.array([-1, -2,  0,  2,  1], np.double)
        S = np.array([-1,  0,  2,  0, -1], np.double)
    else:
        L = np.array([ 1,  6,  15,  20,  15,  6,  1], np.double)
        E = np.array([-1, -4,  -5,   0,   5,  4,  1], np.double)
        S = np.array([-1, -2,   1,   4,   1, -2, -1], np.double)
    oneskernel = np.ones((l,l), np.double)    
    kernels = np.zeros((l,l,9), np.double)
    kernels[:,:,0] = np.multiply(L.reshape(-1,1),L) # LL kernel
    kernels[:,:,1] = np.multiply(L.reshape(-1,1),E) # LE kernel
    kernels[:,:,2] = np.multiply(L.reshape(-1,1),S) # LS kernel
    kernels[:,:,3] = np.multiply(E.reshape(-1,1),L) # EL kernel
    kernels[:,:,4] = np.multiply(E.reshape(-1,1),E) # EE kernel
    kernels[:,:,5] = np.multiply(E.reshape(-1,1),S) # ES kernel
    kernels[:,:,6] = np.multiply(S.reshape(-1,1),L) # SL kernel
    kernels[:,:,7] = np.multiply(S.reshape(-1,1),E) # SE kernel
    kernels[:,:,8] = np.multiply(S.reshape(-1,1),S) # SS kernel
    
    # 4) Get mask where convolution should be performed
    mask_c = _image_xor(mask)
    mask_conv = signal.convolve2d(mask_c, oneskernel,'valid')
    mask_conv = np.abs(np.sign(mask_conv)-1)
        
    # 5) Calculate energy of each convolved image with each kernel: total 9
    energy = np.zeros(9,np.double)   
    area = sum(sum(mask_conv))          
    for i in range(9):
        f_conv = signal.convolve2d(f, kernels[:,:,i], mode='valid')
        f_conv = np.multiply(f_conv,mask_conv)     
        f_conv_mean = sum(sum(f_conv)) / area
        energy[i] = np.sqrt(sum(sum(np.multiply((f_conv-f_conv_mean)**2,mask_conv)))/area)
           
    # 6) Calculate features
    features = np.zeros(6,np.double) 
    features[0] = energy[0]
    features[1] = energy[4]
    features[2] = energy[8]
    features[3] = (energy[1]+energy[3])/2
    features[4] = (energy[5]+energy[7])/2
    features[5] = (energy[2]+energy[6])/2
        
    return features, labels





from .grayscale_morphological_analysis import grayscale_morphology_features, plot_pdf_cdf
from .multilevel_binary_morphological_analysis import multilevel_binary_morphology_features, plot_pdfs_cdfs

__all__ = ['grayscale_morphology_features','plot_pdf_cdf',
           'multilevel_binary_morphology_features','plot_pdfs_cdfs']


# -*- coding: utf-8 -*-
"""
==============================================================================
@author: Nikolaos Giakoumoglou
@date: Wed May 12 17:37:03 2021
@reference: Chua, Automatic indentification of epilepsy by hos and power spectrum parameters using eeg signals
            Chua, Application of Higher Order Spectra to Identify Epileptic eeg
            Acharya, Automatic identification of epileptic eeg singal susing nonlinear parameters
            Acharya, Application of higher order spectra for the identification of diabetes retinopathy stages
==============================================================================            
"""

import numpy as np
from skimage.transform import radon
import matplotlib.pyplot as plt
import warnings
from .bispectrum import _bispectrum  

def _entropy(x):
    return -np.multiply(x, np.log(x+1e-16)).sum()

def hos_features(f, th=[135,140]):
    '''
    Parameters
    ----------
    f : numpy ndarray
        Image of dimensions N1 x N2.
    th : list, optional
        Angle to calculate Radon Transform. The default is [135,140].

    Returns
    -------
    features : numpy ndarray
        Entropy of bispectrum of radeon transform of image for each angle in theta.
    labels : list
        Labels of features.
    '''
    
    warnings.filterwarnings("ignore")
    f = f.astype(np.float32)
    th = np.array([th]).reshape(-1)
    N1, N2 = f.shape
    
    radon_transform = radon(f, theta=th)
    
    entropy = []
    labels = ['HOS_'+str(t)+'_degrees' for t in th]
    for i in range(th.shape[0]):
        B, _ = _bispectrum(radon_transform[:,i])
        p = abs(B) / (sum(sum(abs(B))))
        e = _entropy(p)
        entropy.append(e)
        
    warnings.filterwarnings("default")
    return np.array(entropy).reshape(-1), labels

def plot_sinogram(f, name=''):
    if name != '':
        name = '('+name+')'
    warnings.filterwarnings("ignore")
    th = np.linspace(0., 180., max(f.shape), endpoint=False)
    sinogram = radon(f, theta=th)
    dx, dy = 0.5 * 180.0 / max(f.shape), 0.5 / sinogram.shape[0]
    plt.imshow(sinogram, cmap=plt.cm.Greys_r,
           extent=(-dx, 180.0 + dx, -dy, sinogram.shape[0] + dy),
           aspect='auto')
    plt.title('Sinogram '+name)
    warnings.filterwarnings("default")



# -*- coding: utf-8 -*-
"""
==============================================================================
@author: Nikolaos Giakoumoglou
@date: Thu May 20 18:24:36 2021
@reference: Hamilton, Fast automated cell phenotype image classification
==============================================================================
Threshold Adjacency Statistis (TAS)
==============================================================================
Inputs:
    - f:         image of dimensions N1 x N2
Outputs:
    - features:  feature values
==============================================================================
"""
import mahotas

def tas_features(f):
    '''
    Parameters
    ----------
    f : numpy ndarray
        Image of dimensions N1 x N2.

    Returns
    -------
    features : numpy ndarray
        Feature values.
    labels : list
        Labels of features.
    '''    

    features = mahotas.features.pftas(f)
    labels = ['TAS' + str(i) for i in range(features.shape[0])]
    return features, labels


# -*- coding: utf-8 -*-
"""
==============================================================================
@author: Nikolaos Giakoumoglou
@date: Wed May 12 17:37:03 2021
@reference: Chua, Automatic indentification of epilepsy by hos and power spectrum parameters using eeg signals
            Chua, Application of Higher Order Spectra to Identify Epileptic eeg
            Acharya, Automatic identification of epileptic eeg singal susing nonlinear parameters
            Acharya, Application of higher order spectra for the identification of diabetes retinopathy stages
==============================================================================
"""

from scipy import ndimage
import numpy as np
import matplotlib.pyplot as plt
import math
from .bispectrum import _bispectrum


def _pad_image_2(f):
    '''
    Parameters
    ----------
    f : numpy ndarray
        Image of dimensions N1 x N2.

    Returns
    -------
    f2 : numpy ndarray
        Padded image by N2 x N1.
    '''
    N1, N2 = f.shape
    N1_deficit = N2+2
    N2_deficit = N1+2
    f2 = np.pad(f, ((math.floor(N1_deficit/2),N1_deficit-math.floor(N1_deficit/2)), 
                    (math.floor(N2_deficit/2),N2_deficit-math.floor(N2_deficit/2))), 
                mode='constant')
    return f2


def discrete_radon_transform(f, theta, remove_zeros=False):
    '''
    Parameters
    ----------
    f : numpy ndarray
        Image of dimensions N1 x N2.
    th : list, optional
        Angle to calculate Radon Transform.
    remove_zeros : bool, optional
        Remove zeros from Radon transform. The default is False.

    Returns
    -------
    res : numpy ndarray
        Radon transform of f.
    '''
    f2 = _pad_image_2(f) # so i can rotate without loosing information
    res = np.zeros((len(f2[0]), len(theta)), dtype='float64')
    for i,th in enumerate(theta):
        rotation = ndimage.rotate(f2, th, reshape=False).astype('float64')
        res[:,i] = sum(rotation).reshape(-1)
    if remove_zeros == True:
        res = res[~np.all(abs(res) < 1e-16, axis=1)]
    return res


def _entropy(x):
    return -np.multiply(x, np.log(x+1e-16)).sum()


def hos_features(f, th=[135,140]):
    '''
    Parameters
    ----------
    f : numpy ndarray
        Image of dimensions N1 x N2.
    th : list, optional
        Angle to calculate Radon Transform. The default is [135,140].

    Returns
    -------
    features : numpy ndarray
        Entropy of bispectrum of radeon transform of image for each angle in theta.
    labels : list
        Labels of features.
    '''
        
    f = f.astype(np.float32)   
    radon_transform = discrete_radon_transform(f, th, remove_zeros=True)
    
    labels = ['HOS_'+str(th)+'_degrees' for th in th]
    
    entropy = []
    for i in range(len(th)):
        B, _ = _bispectrum(radon_transform[:,i])
        p = abs(B) / abs(B).sum()
        e = _entropy(p)
        entropy.append(e)
    
    return np.array(entropy).reshape(-1), labels


def plot_sinogram(f, name=''):
    if name != '':
        name = '('+name+')'
    theta = [i for i in range(180)]
    sinogram = discrete_radon_transform(f, theta, True)
    
    dx, dy = 0.5 * 180.0 / max(f.shape), 0.5 / sinogram.shape[0]
    plt.imshow(sinogram, cmap='gray',
           extent=(-dx, 180.0 + dx, -dy, sinogram.shape[0] + dy),
           aspect='auto')
    plt.title('Sinogram '+name)
    plt.show()


# -*- coding: utf-8 -*-
"""
==============================================================================
@author: Nikolaos Giakoumoglou
@date: Fri May 14 10:33:48 2021
==============================================================================
"""
import numpy as np
import matplotlib.pyplot as plt
from scipy.spatial import distance

def correlogram(f, mask, bins_digitize = 32, bins_hist = 32, flatten=False):
    ''' 
    Parameters
    ----------
    f : numpy ndarray
        Image of dimensions N1 x N2.
    mask : numpy ndarray
        Mask image N1 x N2 with 1 if pixels belongs to ROI, 0 else. Give None
        if you want to consider ROI the whole image.
    bins_digitize : int, optional
         Number of bins for discrete distances and thetas. The default is 32.
    bins_hist : int, optional
        Number of bins for histogram. The default is 32.
    flatten : bool, optional
        Return correlogram as 1d array if True or 2d array if False. The 
        default is False.

    Returns
    -------
    Hd : numpy ndarray
        Correlogram for distance.
    Ht : numpy ndarray
        Correlogram for angles.
    labels : list
        Labels of features.
    '''
    
    if mask is None:
        mask = np.ones(f.shape) 
    
    Ng = 256
    
    # Step 1: Find center pixel
    N1, N2 = f.shape
    c = [np.floor(N1/2), np.floor(N2/2)]
    
    # Step 2.1: Find distance of each pixel from center pixels
    D = np.zeros((N1,N2), np.double) * np.nan
    for i in range(N1):
        for j in range(N2):
            D[i,j] = distance.euclidean(c, [i,j])
    D[mask == 0] = 0
    D /= D.max() # normalise to [0,1]
    D = np.digitize(D, bins=np.arange(0,1,1/bins_digitize), 
                    right=False).astype(np.float)
    D[mask == 0] = np.nan
    
    # Step 2.2 Make histogram
    Hd = np.zeros((bins_digitize,bins_hist), np.double)
    for b in range(bins_digitize):
        Hd[b,:] = np.histogram(f[D==(b+1)], bins=bins_hist, range=(0,Ng-1))[0]
        
        
    # Step 3.1: Find angle of each pixel from center pixel
    T = np.zeros((N1,N2), np.double) * np.nan
    for i in range(N1):
        for j in range(N2):
            x = j - c[1]
            y = -(i - c[0])
            angle = np.arctan2(y,x)
            T[i,j] = np.degrees(angle)
            
    # angles between -180 and 180 degrees
    T = np.digitize(T, bins=np.arange(-180,180,360/bins_digitize), 
                    right=False).astype(np.float)
    T[mask == 0] = np.nan
    
    # Step 3.2: Make histogram
    Ht = np.zeros((bins_digitize,bins_hist), np.double)
    for b in range(bins_digitize):
        Ht[b,:] = np.histogram(f[T==(b+1)], bins=bins_hist, range=(0,Ng-1))[0]

    # Step 4: Create label
    labels = []
    for i in range(bins_digitize):
        for j in range(bins_hist):
            labels.append('Correlogram_'+str(i)+'_'+str(j))
    
    # Step 5: Return flatten or 2d array
    if (flatten == True):
        return Hd.flatten(), Ht.flatten(), labels
    else:
        return Hd, Ht, labels

def plot_correlogram(f, mask, bins_digitize = 32, bins_hist = 32, name=''):
    Hd, Ht, _ = correlogram(f, mask, bins_digitize, bins_hist, False)
    if name != '':
        name = '('+name+')'
    fig, ax = plt.subplots()
    ax.imshow(Hd, cmap='hot')
    ax.set_title('Correlogram '+str(Hd.shape[0])+'x'+str(Hd.shape[1])+': Distances'
                 +' '+name)
    
    fig, ax = plt.subplots()
    ax.imshow(Ht, cmap='hot')
    ax.set_title('Correlogram '+str(Ht.shape[0])+'x'+str(Ht.shape[1])+': Angles'
                 +' '+name)
    plt.show()




# -*- coding: utf-8 -*-
"""
==============================================================================
@author: Nikolaos Giakoumoglou
@date: Thu May  6 21:59:57 2021
==============================================================================
"""

import numpy as np

def fos(f, mask):
    '''
    Parameters
    ----------
    f : numpy ndarray
        Image of dimensions N1 x N2.
    mask : numpy ndarray
        Mask image N1 x N2 with 1 if pixels belongs to ROI, 0 else. Give None
        if you want to consider ROI the whole image.

    Returns
    -------
    features : numpy ndarray
        1)Mean, 2)Variance, 3)Median (50-Percentile), 4)Mode, 
        5)Skewness, 6)Kurtosis, 7)Energy, 8)Entropy, 
        9)Minimal Gray Level, 10)Maximal Gray Level, 
        11)Coefficient of Variation, 12,13,14,15)10,25,75,90-
        Percentile, 16)Histogram width
    labels : list
        Labels of features.
    '''
    if mask is None:
        mask = np.ones(f.shape)
    
    # 1) Labels
    labels = ["FOS_Mean","FOS_Variance","FOS_Median","FOS_Mode","FOS_Skewness",
              "FOS_Kurtosis","FOS_Energy","FOS_Entropy","FOS_MinimalGrayLevel",
              "FOS_MaximalGrayLevel","FOS_CoefficientOfVariation",
              "FOS_10Percentile","FOS_25Percentile","FOS_75Percentile",
              "FOS_90Percentile","FOS_HistogramWidth"]
    
    # 2) Parameters
    f  = f.astype(np.uint8)
    mask = mask.astype(np.uint8)
    level_min = 0
    level_max = 255
    Ng = (level_max - level_min) + 1
    bins = Ng
    
    # 3) Calculate Histogram H inside ROI
    f_ravel = f.ravel() 
    mask_ravel = mask.ravel() 
    roi = f_ravel[mask_ravel.astype(bool)] 
    H = np.histogram(roi, bins=bins, range=[level_min, level_max], density=True)[0]
    
    # 4) Calculate Features
    features = np.zeros(16,np.double)  
    i = np.arange(0,bins)
    features[0] = np.dot(i,H)
    features[1] = sum(np.multiply(((i-features[0])**2),H))
    features[2] = np.percentile(roi,50) 
    features[3] = np.argmax(H)
    features[4] = sum(np.multiply(((i-features[0])**3),H))/(np.sqrt(features[1])**3)
    features[5] = sum(np.multiply(((i-features[0])**4),H))/(np.sqrt(features[1])**4)
    features[6] = sum(np.multiply(H,H))
    features[7] = -sum(np.multiply(H,np.log(H+1e-16)))
    features[8] = min(roi)
    features[9] = max(roi)
    features[10] = np.sqrt(features[2]) / features[0]
    features[11] = np.percentile(roi,10) 
    features[12] = np.percentile(roi,25)  
    features[13] = np.percentile(roi,75) 
    features[14] = np.percentile(roi,90) 
    features[15] = features[14] - features[11]
    
    return features, labels



# -*- coding: utf-8 -*-
"""
==============================================================================
@author: Nikolaos Giakoumoglou
@date: Tue May 11 13:20:14 2021
@reference: Maragos, Pattern Spectrum and Multiscale Shape Representation
            Maragos, Threshold Superposition in Morphological Image Analysis Systems
==============================================================================
"""
import numpy as np
import matplotlib.pyplot as plt
from skimage import morphology

def _get_binary_images(img, mask, thresholds=[25,50]): 
    ''' 
    Parameters
    ----------
    img : numpy ndarray
        Image of dimensions N1 x N2.
    mask : numpy ndarray
        Mask image N1 x N2 with 1 if pixels belongs to ROI, 0 else. Give None
        if you want to consider ROI the whole image.
    thresholds: list
        Thresholds to get the 3 binary images. The default is [25, 50].

    Returns
    -------
    L : numpy ndarray
        Low image.
    M : numpy ndarray
        Medium image.
    H : numpy ndarray
        High image.
    '''
    L = np.zeros(img.shape,np.uint8)
    M = np.zeros(img.shape,np.uint8)
    H = np.zeros(img.shape,np.uint8)   
    for i in range(img.shape[0]):
        for j in range(img.shape[1]):
            if mask[i,j] == 1:
                if img[i,j] < thresholds[0]:
                    L[i,j] = 255
                elif img[i,j] > thresholds[1]:
                    H[i,j] = 255
                else:
                    M[i,j] = 255
    return L, M, H

def _opening_SP(X,B,n):
    ''' 
    Parameters
    ----------
    X : numpy ndarray
        Image of dimensions N1 x N2.
    B : numpy ndarray
        Structural element/pattern/kernel
    n : np.array
        Number of scales.

    Returns
    -------
    out : numpy ndarray
        Multiscale set-processing opening (X o nB) at scale n.
    '''
    out = X.copy()
    for i in range(n):
        out = morphology.binary_erosion(out, B)
    for i in range(n):
        out = morphology.binary_dilation(out,B)
    return out.astype(np.uint8)

def _pattern_spectrum(X,B,n):
    ''' 
    Parameters
    ----------
    X : numpy ndarray
        Image of dimensions N1 x N2.
    B : numpy ndarray
        Structural element/pattern/kernel
    n : np.array
        Number of scales.

    Returns
    -------
    out : numpy ndarray
        Pattern spectrum PS(X,B,n) = A[X o nB - X o (n+1)B] 
    '''    
    ps = _opening_SP(X,B,n) - _opening_SP(X,B,n+1)
    return np.count_nonzero(ps)

def _multilevel_binary_morphological_analysis(X, B, N):    
    ''' 
    Parameters
    ----------
    X : numpy ndarray
        Image of dimensions N1 x N2.
    B: numpy ndarray
        Structural element/pattern/kernel
    N : np.array
        Maximum number of scales.

    Returns
    -------
    pdf : numpy ndarray
        Probability density function (pdf) of pattern spectrum.
    cdf : numpy ndarray
        Cumulative density function (cdf) of pattern spectrum.
    '''    
    ps = np.zeros(N,np.double)
    for n in range(N):
        ps[n] = _pattern_spectrum(X,B,n)
    pdf = ps / np.count_nonzero(X)
    cdf = np.cumsum(pdf)
    return pdf, cdf

def multilevel_binary_morphology_features(img, mask, N, thresholds=[25,50]):
    ''' 
    Parameters
    ----------
    img : numpy ndarray
        Image of dimensions N1 x N2.
    mask : numpy ndarray
        Mask image N1 x N2 with 1 if pixels belongs to ROI, 0 else. Give None
        if you want to consider ROI the whole image.
    N : np.array, optional
        Maximum number of scales. The default is 30.
    thresholds: list, optional
        Thresholds to get the 3 binary images. The default is [25, 50].

    Returns
    -------
    pdf_L : numpy ndarray
        Probability density function (pdf) of pattern spectrum for image L.
    pdf_M : numpy ndarray
        Probability density function (pdf) of pattern spectrum for image M.
    pdf_H : numpy ndarray
        Probability density function (pdf) of pattern spectrum for image H.
    cdf_L : numpy ndarray
        Cumulative density function (cdf) of pattern spectrum for image L.
    cdf_M : numpy ndarray
        Cumulative density function (cdf) of pattern spectrum for image M.
    cdf_H : numpy ndarray
        Cumulative density function (cdf) of pattern spectrum for image H.
    '''    
    if mask is None:
        mask = np.ones(img.shape)
    img = img.astype(np.uint8) # grayscale image
    mask = mask.astype(np.uint8)
    L, M, H = _get_binary_images(img, mask, thresholds=thresholds)
    kernel = np.ones((3,3), np.uint8) # kernel/structuring element
    kernel[0,0], kernel[2,2], kernel[0,2], kernel[2,0] = 0, 0, 0, 0 # cross '+'
    pdfs, cdfs = np.zeros((N,3),np.double), np.zeros((N,3),np.double)
    for i, a in enumerate([L, M, H]):
        pdfs[:,i], cdfs[:,i] = _multilevel_binary_morphological_analysis(a,kernel,N)    
    pdf_L, pdf_M, pdf_H = np.array_split(pdfs, 3, axis=1)
    cdf_L, cdf_M, cdf_H = np.array_split(cdfs, 3, axis=1)
    return pdf_L.flatten(), pdf_M.flatten(), pdf_H.flatten(), cdf_L.flatten(), cdf_M.flatten(), cdf_H.flatten()

def plot_pdfs_cdfs(pdf_L, pdf_M, pdf_H, cdf_L, cdf_M, cdf_H, name=''):
    
    if name != '':
        name = '('+name+')'
        
    fig, axs = plt.subplots(3,2)
    fig.suptitle('Plaque Multilevel Binary Morphological Features ' + str(name))
    
    axs[0,0].plot(pdf_L)
    axs[0,1].plot(cdf_L)
    axs[0,0].set_title('pdf low image')
    axs[0,1].set_title('cdf low image')
    
    axs[1,0].plot(pdf_M)
    axs[1,1].plot(cdf_M)
    axs[1,0].set_title('pdf mid image')
    axs[1,1].set_title('cdf mid image')
    
    axs[2,0].plot(pdf_H)
    axs[2,1].plot(cdf_H)
    axs[2,0].set_title('pdf high image')
    axs[2,1].set_title('cdf high image')


# -*- coding: utf-8 -*-
"""
==============================================================================
@author: Kyriacos Constantinou
@author: Ioannis Constantinou,
@author: Marios S. Pattichis
@author: Constantinos S. Pattichis
@reference: Murray, An AM-FM model for Motion Estimation in Atherosclerotic Plaque Videos
            Murray, Multiscale AMFM Demodulation and Image Reconstruction methods with Improved Accuracy 
            Pattichis, Medical Image Analysis Using AM-FM Models and Methods
==============================================================================
"""

import numpy as np
from  scipy import signal
import warnings

def _gabor_kernel_2D(theta, lamda, gamma, bandwidth, phase, overlapIndex):
    qFactor = (1/np.pi) * np.sqrt( (np.log(overlapIndex)/2) ) *  \
                    ( (2**bandwidth + 1) / (2**bandwidth - 1) )
    sigma = lamda*qFactor
    n = np.ceil(4*sigma)
    [x,y] = np.mgrid[-n:(n+2),-n:(n+2)]
    xTheta = x * np.cos(theta) + y * np.sin(theta)
    yTheta = -x * np.sin(theta) + y * np.cos(theta)
    gaussian = np.exp(-(( xTheta**2) + gamma**2.* (yTheta**2))/(2*sigma**2))
    res = gaussian * np.cos(2*np.pi*xTheta/lamda +  phase)
    maxFft = abs(np.fft.fft2(res)).max()
    normalize = np.fft.fft2(res)/maxFft
    result = np.real(np.fft.ifft2(normalize))
    return result, sigma
    
def _gaussian_function(f0, s0, overlapIndex):
    over = np.sqrt(2*np.log(1/overlapIndex))
    sigma = s0*over/(s0*f0 - over)
    n = np.ceil(2*sigma)
    [x,y] = np.mgrid[-n:(n+2),-n:(n+2)]
    res = np.exp(-1/2*(x**2 + y**2)/sigma**2)
    res = res / res.sum()
    return res
 
def _filterbank():
    '''
    Returns
    -------
    filter : list
        List of 41 filters for AM-FM multi-scale analysis.
    '''
    
    lamda0 = 2
    orientations = 8
    scales = 5
    gamma = 1
    phase = 0
    bandwidth = 1
    overlapIndex = 0.5
    offset = 0
    theta = np.arange(offset,np.pi - np.pi/orientations + offset + 
                             np.pi/orientations, (np.pi/orientations))
    lamda = lamda0
    filters = []
    
    for sc_index in range(scales, 0, -1):
        lamda0 = lamda
        for th in range(theta.shape[0]):
            result, sig = _gabor_kernel_2D(theta[th], lamda, gamma, bandwidth, phase, 1/overlapIndex);
            filters .append(result)
        lamda = lamda0 * 2**bandwidth
    # Add DC Filter
    f1      = 2*np.pi/lamda
    result  = _gaussian_function(f1, sig, overlapIndex)
    filters.append(result)
    return filters

def _calculate_amfm(f):
    '''
    Parameters
    ----------
    f : numpy ndarray
        Image of dimensions N1 x N2.

    Returns
    -------
    IA : numpy ndarray
        instantaneous amplitude (a_n).
    IP : numpy ndarray
        instanteneous phase (φ_n).
    IFx : numpy ndarray
        instantanteous frequency (grad φ1_n).
    IFy : numpy ndarray
        instantanteous frequency (grad φ2_n).
    '''
    N1, N2 = f.shape
    IA = np.abs(f)
    IP = np.angle(f)
    IANorm = np.divide(f, IA+1e-16)
    IFx = np.zeros((N1,N2), np.double)
    IFy = np.zeros((N1,N2), np.double)
    for i in range(1,N1-1):
        for j in range(1,N2-1):
            IFx[i,j] = np.abs(np.arccos(np.real((IANorm[i+1,j]+IANorm[i-1,j]) / (2*IANorm[i,j]))))
            IFy[i,j] = np.abs(np.arccos(np.real((IANorm[i,j+1]+IANorm[i,j-1]) / (2*IANorm[i,j]))))
    return IA, IP, IFx, IFy

def _dca(band):
    '''
    Parameters
    ----------
    band : list
        The band. A list of IA, IP, IFx, IFy.

    Returns
    -------
    IA : list
        Max instantaneous amplitude for given band.
    IP : list
        Max instanteneous phase for given band.
    IFx : list
        Max instantanteous frequency for given band.
    IFy : list
        Max instantanteous frequency for given band.
    '''
    
    IA = np.zeros(band[0][0].shape)
    IP = np.zeros(band[0][1].shape)
    IFx = np.zeros(band[0][2].shape)
    IFy = np.zeros(band[0][3].shape)
    w, h = band[0][0].shape
    
    for i in range(w):
        for j in range(h):
            pos = 0
            temp = band[pos][0][i,j] # band[pos].IA[i,j]
            for l in range(len(band)):
                if temp < band[l][0][i,j]: # band[l].IA[i,j]
                    pos = l
                    temp = band[l][0][i,j]
            IA[i,j] = temp
            IP[i,j] = band[pos][1][i,j]
            IFx[i,j] = band[pos][2][i,j]
            IFy[i,j] = band[pos][3][i,j]
    
    return IA, IP, IFx, IFy
    
def amfm_features(f, bins=32):
    '''
    Parameters
    ----------
    f : numpy ndarray
        Image of dimensions N1 x N2.
    bins: int, optional
        Bins for the calculated histogram. The default is 32.

    Returns
    -------
    features : numpy ndarray
        Histogram of IA, IP, IFx, IFy as a concatenated vector.
    labels : list
        Labels of features.
    '''
    

    warnings.simplefilter(action='ignore', category=RuntimeWarning)
    AMFM = []
    filters = _filterbank()
    f_hilbert = signal.hilbert(f)
    
    #mask_c = _image_xor(mask)
    #mask_conv = []
    #for filtre in filters:
    #    oneskernel = np.ones(filtre.shape)
    #    temp = signal.convolve2d(mask_c, oneskernel,'same')
    #    temp = np.abs(np.sign(temp)-1)
    #    mask_conv.append(temp)
        
    for i, filtre in enumerate(filters):
        f_filtered = signal.convolve2d(f_hilbert, np.rot90(filtre), mode='same', boundary='fill', fillvalue=0)
        #f_filtered = f_filtered * mask_conv[i]
        IA, IP, IFx, IFy = _calculate_amfm(f_filtered)
        IA = np.nan_to_num(IA)
        IP = np.nan_to_num(IP)
        IFx = np.nan_to_num(IFx)
        IFy = np.nan_to_num(IFy)
        AMFM.append([IA, IP, IFx, IFy])
     
    # Access like this: band[i][0] for IA, band[i][1] for IP,
    # band[i][2] for IFx and band[i][3] for IFy
    high = []
    med = []
    low = []
    dc = []
    for i in range(len(filters)):
        if (i <= 7):
            high.append(AMFM[i])
        elif (i<=23):
            med.append(AMFM[i])
        elif (i<=39):
            low.append(AMFM[i])
        else:
            dc.append(AMFM[i])
    
    IAl, IPl, IFxl, IFyl = _dca(low)
    IAl = (IAl > np.percentile(IAl,50)).astype(np.float64) * IAl
    reconstructionImgDCAl = np.real(IAl * np.cos(IPl))
    H1 = np.histogram(reconstructionImgDCAl, bins=bins, density=True)[0]
    
    IAm, IPm, IFxm, IFym = _dca(med)
    IAm = (IAm > np.percentile(IAl,50)).astype(np.float64) * IAm
    reconstructionImgDCAm = np.real(IAm * np.cos(IPm))
    H2 = np.histogram(reconstructionImgDCAm, bins=bins, density=True)[0]
    
    IAh, IPh, IFxh, IFyh = _dca(high)
    IAh = (IAh > np.percentile(IAl,50)).astype(np.float64) * IAh
    reconstructionImgDCAh = np.real(IAh * np.cos(IPh))
    H3 = np.histogram(reconstructionImgDCAh, bins=bins, density=True)[0]
    
    IAdc, IPdc, IFxdc, IFydc = _dca(dc)
    reconstructionImgDCAdc = np.real(IAdc * np.cos(IPdc))
    H4 = np.histogram(reconstructionImgDCAdc, bins=bins, density=True)[0]
    
    features = np.concatenate([H1, H2, H3, H4])
    labels = []
    labels.append(['AMFM_low'+str(i) for i in range(32)])
    labels.append(['AMFM_med'+str(i) for i in range(32)])
    labels.append(['AMFM_high'+str(i) for i in range(32)])
    labels.append(['AMFM_dc'+str(i) for i in range(32)])
    labels = [item for sublist in labels for item in sublist]
    
    warnings.simplefilter(action='default', category=RuntimeWarning)
    
    return features, labels
    
def plotAMFM(f):
    pass
    


"""
@author: github.com/synergetics/spectrum
"""

from __future__ import division
import numpy as np
from scipy.linalg import hankel
from scipy.signal import convolve2d
import matplotlib.pyplot as plt

def _nextpow2(num):
  npow = 2
  while npow <= num:
      npow = npow * 2
  return npow


def _flat_eq(x, y):
  z = x.reshape(1, -1)
  z = y
  return z.reshape(x.shape)


def _make_arr(arrs, axis=0):
  a = []
  ctr = 0
  for x in arrs:
    if len(np.shape(x)) == 0:
      a.append(np.array([[x]]))
    elif len(np.shape(x)) == 1:
      a.append(np.array([x]))
    else:
      a.append(x)
    ctr += 1
  return np.concatenate(a, axis)


def shape(o, n):
  s = o.shape
  if len(s) < n:
    x = tuple(np.ones(n-len(s)))
    return s + x
  else:
    return s

def _bispectrum(y, nlag=None, nsamp=None, overlap=None,
  flag='biased', nfft=None, wind=None):
  """
  Parameters:
    y       - data vector or time-series
    nlag    - number of lags to compute [must be specified]
    segsamp - samples per segment    [default: row dimension of y]
    overlap - percentage overlap     [default = 0]
    flag    - 'biased' or 'unbiased' [default is 'unbiased']
    nfft    - FFT length to use      [default = 128]
    wind    - window function to apply:
              if wind=0, the Parzen window is applied (default)
              otherwise the hexagonal window with unity values is applied.
  Output:
    Bspec   - estimated bispectrum  it is an nfft x nfft array
              with origin at the center, and axes pointing down and to the right
    waxis   - frequency-domain axis associated with the bispectrum.
            - the i-th row (or column) of Bspec corresponds to f1 (or f2)
              value of waxis(i).
  """
 
  ly = y.shape[0]
  nrecs = 1
  
  if not overlap: overlap = 0
  overlap = min(99, max(overlap,0))
  if nrecs > 1: overlap = 0
  if not nsamp: nsamp = ly
  if nsamp > ly or nsamp <= 0: nsamp = ly
  if not 'flag': flag = 'biased'
  if not nfft: nfft = 128
  if not wind: wind = 0

  nlag = nsamp-1
  if nfft < 2*nlag+1:
    nfft = 2^_nextpow2(nsamp)


  # create the lag window
  Bspec = np.zeros([nfft, nfft])
  if wind == 0:
    indx = np.array([range(1,nlag+1)]).T
    window = _make_arr((1, np.sin(np.pi*indx/nlag) / (np.pi*indx/nlag)), axis=0)
  else:
    window = np.ones([nlag+1,1])
  window = _make_arr((window, np.zeros([nlag,1])), axis=0)

  # cumulants in non-redundant region
  overlap  = np.fix(nsamp * overlap / 100)
  nadvance = nsamp - overlap
  nrecord  = np.fix((ly*nrecs - overlap) / nadvance)

  c3 = np.zeros([nlag+1,nlag+1])
  ind = np.arange(nsamp)
  y = y.ravel(order='F')

  s = 0
  for k in range(nrecord.astype('i')):
    x = y[ind].ravel(order='F')
    x = x - np.mean(x)
    ind = ind + int(nadvance)

    for j in range(nlag+1):
      z = x[range(nsamp-j)] * x[range(j, nsamp)]
      for i in range(j,nlag+1):
        Sum = np.dot(z[range(nsamp-i)].T, x[range(i,nsamp)])
        if flag == 'biased': Sum = Sum/nsamp
        else: Sum = Sum / (nsamp-i)
        c3[i,j] = c3[i,j] + Sum

  c3 = c3 / nrecord

  # cumulants elsewhere by symmetry
  c3 = c3 + np.tril(c3,-1).T   # complete I quadrant
  c31 = c3[1:nlag+1, 1:nlag+1]
  c32 = np.zeros([nlag, nlag])
  c33 = np.zeros([nlag, nlag])
  c34 = np.zeros([nlag, nlag])
  for i in range(nlag):
    x = c31[i:nlag, i]
    c32[nlag-1-i,0:nlag-i] = x.T
    c34[0:nlag-i, nlag-1-i] = x
    if i+1 < nlag:
      x = np.flipud(x[1:len(x)])
      c33 = c33 + np.diag(x,i+1) + np.diag(x,-(i+1))

  c33  = c33 + np.diag(c3[0, nlag:0:-1])

  cmat = _make_arr(
    (_make_arr((c33, c32, np.zeros([nlag,1])), axis=1),
    _make_arr((_make_arr((c34, np.zeros([1,nlag])), axis=0), c3), axis=1)),
    axis=0
  )

  # apply lag-domain window
  wcmat = cmat
  if wind != -1:
    indx = np.arange(-1*nlag, nlag+1).T
    window = window.reshape(-1,1)
    for k in range(-nlag, nlag+1):
      wcmat[:, k+nlag] = (cmat[:, k+nlag].reshape(-1,1) * \
              window[abs(indx-k)] * \
              window[abs(indx)] * \
              window[abs(k)]).reshape(-1,)


  # compute 2d-fft, and shift and rotate for proper orientation
  Bspec = np.fft.fft2(wcmat, (nfft, nfft))
  Bspec = np.fft.fftshift(Bspec) # axes d and r; orig at ctr


  if nfft%2 == 0:
    waxis = np.transpose(np.arange(-1*nfft/2, nfft/2)) / nfft
  else:
    waxis = np.transpose(np.arange(-1*(nfft-1)/2, (nfft-1)/2+1)) / nfft

  return (Bspec, waxis)


from .histogram import *
from .textural import *
from .morphological import *
from .multiscale import *
from .other import *

__all__ = ['histogram', 'plot_histogram',
           'multiregion_histogram',
           'correlogram', 'plot_correlogram',
           'fos',
           'glcm_features',
           'glds_features',
           'ngtdm_features',
           'sfm_features',
           'lte_measures',
           'fdta',
           'glrlm_features',
           'fps',
           'shape_parameters',
           'hos_features','plot_sinogram',
           'lbp_features',
           'glszm_features',
           'grayscale_morphology_features','plot_pdf_cdf',
           'multilevel_binary_morphology_features','plot_pdfs_cdfs',
           'fdta',
           'dwt_features',
           'swt_features', 
           'wp_features', 
           'gt_features', 
           'amfm_features',
           'hog_features', 'plot_hog',
           'hu_moments',
           'tas_features',
           'zernikes_moments']




# -*- coding: utf-8 -*-
"""
==============================================================================
@author: Christos Loizou
@date: Sat May  8 09:42:10 2021
@reference: Wu, Texture Features for Classification
==============================================================================
Fractal Dimension Texture Analysis
==============================================================================
Inputs:
    - f:        image of dimensions N1 x N2
    - s:        degree to calculate Hurst coefficients
Outputs:
    - features: Hurst Coefficients
==============================================================================
"""
import numpy as np


def _least(id, k):
    '''
    Estimation of the curve slope using least squares regression, in log-log scale
    '''
    return np.polyfit(np.log10(np.arange(1,k+1)), np.log10(id), deg=1)[0]


def _intensity(f, mask, s):
    '''
    Intensity difference vector with step 
    '''
    n1, n2, cn1, cn2 = 0, 0, 0, 0
    N1, N2 = f.shape
    IDV = np.zeros((s), np.double)
        
    for k in range(1,s+1):
        for x1 in range(0,N1):
            for y1  in range(0,N2-k):
                if (mask[x1,y1] == 1) & (mask[x1,y1+k] == 1):
                    n1 += np.abs(f[x1,y1]-f[x1,y1+k])
                    cn1 += 1
    
        for x2 in range(0,N1-k):
            for y2 in range(0,N2):
                if (mask[x2,y2] == 1) & (mask[x2+k,y2] == 1):
                    n2 += np.abs(f[x2,y2]-f[x2+k,y2])
                    cn2 += 1
        IDV[k-1] = (n1+n2)/(cn1+cn2+1e-16)
          
    return IDV


def _resolution(x, mask, mr, mc):  
    '''
    Multiple resolution feature exctraction
    '''
    nr = (2 ** mr) - 1
    nc = (2 ** mc) - 1  
    nr_int = nr.astype('i')
    nc_int = nc.astype('i')
    res = np.zeros((nr_int,nc_int), np.double)
    res_mask = np.zeros((nr_int,nc_int), np.double)
    for i in range(0,nr_int):
        for j in range(0,nc_int):
            res[i,j]=(x[(2*i),(2*j)]+x[(2*i+1),(2*j)]+x[(2*i),(2*j+1)]+x[(2*i+1),(2*j+1)])/4
            res_mask[i,j]=(mask[(2*i),(2*j)]+mask[(2*i+1),(2*j)]+mask[(2*i),(2*j+1)]+mask[(2*i+1),(2*j+1)])/4
    res_mask[res_mask>1] = 1
    return res, res_mask
    

def fdta(f, mask, s=3):
    '''
    Parameters
    ----------
    f : numpy ndarray
        Image of dimensions N1 x N2.
    mask : numpy ndarray
        Mask image N1 x N2 with 1 if pixels belongs to ROI, 0 else. Give None
        if you want to consider ROI the whole image.
    s : int, optional
        max resolution to calculate Hurst coefficients. The default is 3

    Returns
    -------
    h : numpy ndarray
        Hurst coefficients.
    labels : list
        Labels of h.
    '''
    
    if mask is None:
        mask = np.ones(f.shape)
        
    labels = ["FDTA_HurstCoeff"] * (s+1)
    labels = [label + "_" + str(i+1) for i,label in enumerate(labels)]
    f = np.array(f, np.double)
    N1, N2 = f.shape
    h = np.zeros((s+1), np.double)
    h[s] = 0
    ms = 3
    i = 0
    IDV = _intensity(f,mask,ms)
    h[i] = _least(IDV,ms)
    mr = np.log2(N1)
    mc = np.log2(N2)
    while (i < s):
        i = i + 1
        mr = mr - 1
        mc = mc - 1
        f, mask = _resolution(f,mask,mr,mc)
        IDV = _intensity(f,mask,ms)
        h[i] = _least(IDV,ms)
    return h, labels 



# -*- coding: utf-8 -*-
"""
==============================================================================
@author: Nikolaos Giakoumoglou
@date: Thu May  6 19:10:11 2021
@reference: Weszka, A Comparative Study of Texture Measures for Terrain Classification
==============================================================================
"""

import numpy as np

def glds(f, mask, dx, dy, Ng):
    '''
    Parameters
    ----------
    f : numpy ndarray
        Image of dimensions N1 x N2.
    mask : numpy ndarray
        Mask image N1 x N2 with 1 if pixels belongs to ROI, 0 else
    dx : int
        Orientation in X-coordinate
    dy : int
        Orientation in Y-coordinate
    Ng : int
        Image number of gray values

    Returns
    -------
    f_d : numpy ndarray
    p_d : numpy ndarray
    '''
    
    N1, N2 = f.shape
    
    # Calculate f_d(x,y). If calculation includes pixel outside mask, ignore it.
    f_d = np.zeros((N1,N2), np.double) 
    for x in range(N1):
        for y in range(N2):
            if (x+dx < N1) & (y+dy < N2) & (x+dx >= 0) & (y+dy >= 0) & (mask[x,y]==1):
                f_d[x,y] = abs(f[x,y] - f[x+dx,y+dy])      
            
    # Calculate pd_(i)
    f_d_ravel = f_d.ravel()
    mask_ravel = mask.ravel()
    roi = f_d_ravel[mask_ravel.astype(bool)]
    p_d = np.histogram(roi, bins=Ng, range=(0,Ng-1))[0] # histogram of f_d in ROI
        
    return f_d, p_d
     
def glds_features(f, mask, Dx=[0,1,1,1], Dy=[1,1,0,-1]):
    '''
    Parameters
    ----------
    f : numpy ndarray
        Image of dimensions N1 x N2.
    mask : numpy ndarray
        Mask image N1 x N2 with 1 if pixels belongs to ROI, 0 else.
    Dx : int, optional
        Array with X-coordinates of vectors denoting orientation. The default
        is [0,1,1,1].
    Dy : int, optional
        Array with Y-coordinates of vectors denoting orientation. The default
        is [1,1,0,-1].
        
    Returns
    -------
    features : numpy ndarray
        1) Contrast, 2)Angular Second Moment, 3)Entropy, 4)Mean
    labels : list
        Labels of features.
    '''
    
    if mask is None:
        mask = np.ones(f.shape)
        
    # 1) Labels
    labels =  ["GLDS_Homogeneity","GLDS_Contrast",
               "GLDS_ASM","GLDS_Entopy","GLDS_Mean"]
    
    # 2) Parameters
    f = f.astype(np.double)
    mask = mask.astype(np.double)
    Dx = np.array(Dx)
    Dy = np.array(Dy)
    Ng = 256    
    
    # 3) Loop over Dx, Dy values to calculate feats
    features = []  
    for ii in range(Dx.shape[0]):
        
        dx = Dx[ii]
        dy = Dy[ii]  
        
        f_d, p_d = glds(f, mask, dx, dy, Ng)
            
        feats = np.zeros(5,np.double)
        i = np.arange(Ng)
        i2 = i ** 2
        feats[0] = sum(np.divide(p_d,i2+1))
        feats[1] = sum(np.multiply(p_d, i2))
        feats[2] = sum(np.multiply(p_d,p_d))
        feats[3] = -sum(np.multiply(p_d,np.log(p_d+1e-16)))
        feats[4] = sum(np.multiply(p_d,i))
        features.append(feats) 
      
    # 4) Calculate Features: mean of feats
    features = np.array(features)
    features = features.mean(axis=0)  
        
    return features, labels